import re
from datetime import datetime
from typing import Annotated, ClassVar, Generic, Optional, Type, TypeVar, overload

from pydantic import BaseModel, ConfigDict, Field, model_validator

from pynspd.errors import UnknownLayer
from pynspd.schemas._common import CamelModel

OptProps = TypeVar("OptProps", bound="OptionProperties")
T = TypeVar("T", bound="OptionProperties")


class NspdProperties(CamelModel, Generic[OptProps]):
    """Базовый класс для валидации поля `property` в GeoJSON-объекте из НСПД"""

    category: int
    """ID категории слоя"""

    category_name: str
    """Имя категории слоя"""

    options: OptProps
    """Свойства объекта"""

    system_info: Optional["SystemInfoProperties"] = None
    cadastral_districts_code: Optional[int] = None
    descr: Optional[str] = None
    external_key: Optional[str] = None
    interaction_id: Optional[int] = None
    label: Optional[str] = None
    subcategory: Optional[int] = None
    score: Optional[int] = None

    @overload
    def cast(self, option_def: None = None) -> "NspdProperties[OptionProperties]": ...

    @overload
    def cast(self, option_def: Type[T]) -> "NspdProperties[T]": ...

    def cast(self, option_def: Optional[Type[T]] = None):
        if option_def is None:
            model = NspdProperties[OptionProperties.by_category_id(self.category)]
        else:
            model = NspdProperties[option_def]
        return model.model_validate(self.model_dump(by_alias=True))

    def get_title(self) -> Optional[str]:
        """Попытка найти заголовок карточки в свойствах"""
        possible_titles = (self.options.title_key, "cad_num", "cad_number")
        props = self.options.model_dump(by_alias=True)
        for title in possible_titles:
            if title is not None and title in props:
                return props[title]
        return None


class OptionProperties(BaseModel):
    """Базовый класс для валидации поля `property.options` в GeoJSON-объекте из НСПД"""

    model_config = ConfigDict(
        extra="allow",
        use_attribute_docstrings=True,
    )

    title_key: ClassVar[Optional[str]] = None
    """Имя поля, которое является заголовком для карточки"""

    # TODO: определять объекты без геометрии отдельной схемой
    no_coords: Annotated[bool, Field(alias="geocoderObject")] = False
    objdoc_id: Annotated[Optional[int], Field(alias="objdocId")] = None
    registers_id: Annotated[Optional[int], Field(alias="registersId")] = None

    @model_validator(mode="before")
    @classmethod
    def valid_date(cls, values: dict) -> dict:
        for k, v in values.items():
            if k in cls.model_fields and "datetime.date" in str(
                cls.model_fields[k].annotation
            ):
                if v == "" or v is None:
                    values[k] = None
                elif isinstance(v, str) and re.match(r"\d+\.\d+\.\d+", v):
                    values[k] = datetime.strptime(v, "%d.%m.%Y")
        return values

    @classmethod
    def by_category_id(cls, category_id: int) -> Type["OptionProperties"]:
        """Получение модели по категории"""
        for sub_class in cls.__subclasses__():
            if sub_class.__name__ == f"Options{category_id}":
                return sub_class
        raise UnknownLayer(category_id)

    def model_dump_human_readable(self):
        """Генерация словаря с ключами, аналогичным карточке на сайте"""
        data = self.model_dump()
        alias = {
            k: v.description for k, v in self.model_fields.items() if v.description
        }
        aliased_data = {alias[k]: v for k, v in data.items() if k in alias}
        return aliased_data


class SystemInfoProperties(CamelModel):
    inserted: Optional[datetime]
    inserted_by: Optional[str]
    updated: Optional[datetime]
    updated_by: Optional[str]



import json
from typing import Annotated, Generic, Literal, TypeVar, Union

import geojson_pydantic.geometries as pyd_geom
import pyproj
import shapely.geometry as shape_geom
from pydantic import BaseModel, Field, StringConstraints, model_validator
from shapely import to_geojson
from shapely.geometry.base import BaseGeometry
from shapely.ops import transform

BaseGeomT = TypeVar("BaseGeomT", bound=BaseGeometry)


class CoordinateReferenceSystemProps(BaseModel):
    name: Annotated[str, StringConstraints(pattern=r"EPSG:\d+")]


class CoordinateReferenceSystem(BaseModel):
    type: Literal["name"]
    properties: CoordinateReferenceSystemProps

    @property
    def value(self):
        return int(self.properties.name.split(":")[1])


class ShapeGeometry(BaseModel, Generic[BaseGeomT]):
    @model_validator(mode="before")
    @classmethod
    def force_4326(cls, data: dict) -> dict:
        if data.get("crs") is None:
            return data
        crs = CoordinateReferenceSystem.model_validate(data["crs"])
        if crs.value == 4326:
            return data
        geom = shape_geom.shape(data)
        transformer = pyproj.Transformer.from_crs(
            pyproj.CRS(crs.properties.name),
            pyproj.CRS("EPSG:4326"),
            always_xy=True,
        ).transform
        proj_geom = transform(transformer, geom)
        return json.loads(to_geojson(proj_geom))

    def to_shape(self) -> BaseGeomT:
        """Конвертация в shapely-геометрию"""
        return shape_geom.shape(self)


class Point(pyd_geom.Point, ShapeGeometry[shape_geom.Point]):
    def to_multi_shape(self) -> shape_geom.MultiPoint:
        return shape_geom.MultiPoint([self.to_shape()])


class MultiPoint(pyd_geom.MultiPoint, ShapeGeometry[shape_geom.MultiPoint]):
    def to_multi_shape(self) -> shape_geom.MultiPoint:
        return self.to_shape()


class LineString(pyd_geom.LineString, ShapeGeometry[shape_geom.LineString]):
    def to_multi_shape(self) -> shape_geom.MultiLineString:
        return shape_geom.MultiLineString([self.to_shape()])


class MultiLineString(
    pyd_geom.MultiLineString, ShapeGeometry[shape_geom.MultiLineString]
):
    def to_multi_shape(self) -> shape_geom.MultiLineString:
        return self.to_shape()


class Polygon(pyd_geom.Polygon, ShapeGeometry[shape_geom.Polygon]):
    def to_multi_shape(self) -> shape_geom.MultiPolygon:
        return shape_geom.MultiPolygon([self.to_shape()])


class MultiPolygon(pyd_geom.MultiPolygon, ShapeGeometry[shape_geom.MultiPolygon]):
    def to_multi_shape(self) -> shape_geom.MultiPolygon:
        return self.to_shape()


Geometry = Annotated[
    Union[
        Point,
        MultiPoint,
        LineString,
        MultiLineString,
        Polygon,
        MultiPolygon,
    ],
    Field(discriminator="type"),
]



import httpx
import pytest
from pytest_httpx import HTTPXMock

from pynspd import AsyncNspd


@pytest.mark.asyncio(scope="session")
async def test_handled_proxy_error(httpx_mock: HTTPXMock, async_api: AsyncNspd):
    """Ошибка прокси о невозвожности ресолвить адрес"""
    assert not async_api._dns_resolve
    httpx_mock.add_exception(httpx.ProxyError("Connection not allowed by ruleset"))
    httpx_mock.add_response()
    with pytest.warns():
        await async_api.request("get", "/api")
        assert len(httpx_mock.get_requests()) == 2


@pytest.mark.asyncio(scope="session")
async def test_unhandled_proxy_error(httpx_mock: HTTPXMock, async_api: AsyncNspd):
    """Неизвестная ошибка прокси"""
    httpx_mock.add_exception(httpx.ProxyError("Hello world"))
    with pytest.raises(httpx.ProxyError):
        await async_api.request("get", "/api")
